from markitup import sgr as _sgr


def entry_github(title: str, details: str, pprint: bool = True) -> str:
    output = f"::group::{title}\n{details}\n::endgroup::"
    if pprint:
        print(output, flush=True)
    return output


def entry_console(
    title: str,
    details: str = "",
    seperator_top: str = "="*35,
    seperator_bottom: str = "="*35,
    seperator_title: str = "-"*20,
    pprint: bool = True,
) -> str:
    output = ""
    if seperator_top:
        output += f"{seperator_top}\n"
    output += f"{title}"
    if seperator_title and details:
        output += f"\n{seperator_title}"
    if details:
        output += f"\n{details}"
    if seperator_bottom:
        output += f"\n{seperator_bottom}"
    if pprint:
        print(output, flush=True)
    return output


def h1(
    title: str,
    width: int = 110,
    margin_top: int = 2,
    margin_bottom: int = 1,
    text_styles: str | int | list[str | int] | None = "bold",
    text_color: str | int | tuple[int, int, int] | None = (255, 255, 255),
    background_color: str | int | tuple[int, int, int] | None = (150, 0, 170),
    pprint: bool = True,
) -> str:
    return h(**locals())


def h2(
    title: str,
    width: int = 95,
    margin_top: int = 1,
    margin_bottom: int = 1,
    text_styles: str | int | list[str | int] | None = "bold",
    text_color: str | int | tuple[int, int, int] | None = (255, 255, 255),
    background_color: str | int | tuple[int, int, int] | None = (25, 100, 175),
    pprint: bool = True,
) -> str:
    return h(**locals())


def h3(
    title: str,
    width: int = 80,
    margin_top: int = 1,
    margin_bottom: int = 1,
    text_styles: str | int | list[str | int] | None = "bold",
    text_color: str | int | tuple[int, int, int] | None = (255, 255, 255),
    background_color: str | int | tuple[int, int, int] | None = (100, 160, 0),
    pprint: bool = True,
) -> str:
    return h(**locals())


def h4(
    title: str,
    width: int = 65,
    margin_top: int = 1,
    margin_bottom: int = 0,
    text_styles: str | int | list[str | int] | None = "bold",
    text_color: str | int | tuple[int, int, int] | None = (255, 255, 255),
    background_color: str | int | tuple[int, int, int] | None = (200, 150, 0),
    pprint: bool = True,
) -> str:
    return h(**locals())


def h5(
    title: str,
    width: int = 50,
    margin_top: int = 1,
    margin_bottom: int = 0,
    text_styles: str | int | list[str | int] | None = "bold",
    text_color: str | int | tuple[int, int, int] | None = (255, 255, 255),
    background_color: str | int | tuple[int, int, int] | None = (240, 100, 0),
    pprint: bool = True,
) -> str:
    return h(**locals())


def h6(
    title: str,
    width: int = 35,
    margin_top: int = 1,
    margin_bottom: int = 0,
    text_styles: str | int | list[str | int] | None = "bold",
    text_color: str | int | tuple[int, int, int] | None = (255, 255, 255),
    background_color: str | int | tuple[int, int, int] | None = (220, 0, 35),
    pprint: bool = True,
) -> str:
    return h(**locals())


def h(
    title: str,
    width: int,
    margin_top: int,
    margin_bottom: int,
    text_styles: str | int | list[str | int] | None = None,
    text_color: str | int | tuple[int, int, int] | None = None,
    background_color: str | int | tuple[int, int, int] | None = None,
    pprint: bool = False,
) -> str:
    control_sequence = _sgr.style(
        text_styles=text_styles, text_color=text_color, background_color=background_color
    )
    centered_title = title.center(width)
    heading_box = _sgr.format(text=centered_title, control_sequence=control_sequence)
    margin_top = "\n" * margin_top
    margin_bottom = "\n" * margin_bottom
    output = f"{margin_top}{heading_box}{margin_bottom}"
    if pprint:
        print(output, flush=True)
    return output



from typing import Literal as _Literal
from pathlib import Path as _Path
from markitup import html as _html
from actionman import pprint as _pprint


class Logger:

    def __init__(
        self,
        realtime_output: bool = True,
        github_console: bool = True,
        output_html_filepath: str | _Path | None = "log.html",
        initial_section_level: _Literal[1, 2, 3, 4, 5] = 1,
        h1_kwargs: dict | None = None,
        h2_kwargs: dict | None = None,
        h3_kwargs: dict | None = None,
        h4_kwargs: dict | None = None,
        h5_kwargs: dict | None = None,
        h6_kwargs: dict | None = None,
        symbol_bulletpoint: str = "🔘",
        symbol_success: str = "✅",
        symbol_skip: str = "❎",
        symbol_error: str = "⛔",
        symbol_warning: str = "🚨",
        symbol_attention: str = "❗",
        entry_seperator_top: str = "="*35,
        entry_seperator_bottom: str = "="*35,
        entry_seperator_title: str = "-"*30,
    ):
        self.realtime_output = realtime_output
        self.github_console = github_console
        self.output_html_filepath = _Path(output_html_filepath).resolve() if output_html_filepath else None
        self.section_level = initial_section_level

        self._heading_kwargs = {
            1: h1_kwargs or {},
            2: h2_kwargs or {},
            3: h3_kwargs or {},
            4: h4_kwargs or {},
            5: h5_kwargs or {},
            6: h6_kwargs or {},
        }
        self._heading_pprint = {
            1: _pprint.h1,
            2: _pprint.h2,
            3: _pprint.h3,
            4: _pprint.h4,
            5: _pprint.h5,
            6: _pprint.h6,
        }
        self._bullet = symbol_bulletpoint
        self._status_symbol = {
            "success": symbol_success,
            "skip": symbol_skip,
            "error": symbol_error,
            "warning": symbol_warning,
            "attention": symbol_attention,
        }
        self._entry_seperator_top = entry_seperator_top
        self._entry_seperator_bottom = entry_seperator_bottom
        self._entry_seperator_title = entry_seperator_title

        if self.output_html_filepath:
            self.output_html_filepath.parent.mkdir(parents=True, exist_ok=True)
            self.output_html_filepath.touch(exist_ok=True)

        self._log_console: str = ""
        self._log_html: str = ""
        return

    @property
    def console_log(self):
        return self._log_console

    @property
    def html_log(self):
        return self._log_html

    def section(self, title: str):
        heading_html = _html.h(min(self.section_level + 1, 6), title)
        heading_console = self._heading_pprint[self.section_level](
            title, pprint=False, **self._heading_kwargs[self.section_level]
        )
        self._submit(console=heading_console, file=heading_html)
        self.section_level = min(self.section_level + 1, 6)
        return

    def entry(
        self,
        status: _Literal["info", "debug", "success", "error", "warning", "attention", "skip", "input"],
        title: str,
        summary: str = "",
        details: tuple[str, ...] | list[str] = tuple(),
    ):
        title_full = f"{self._status_symbol[status]} {title}"
        details_console = "\n".join([f"{self._bullet} {detail}" for detail in details])
        details_console_full = (
            f"{summary}\n{details_console}" if summary and details else f"{summary}{details_console}"
        )
        console_entry = _pprint.entry_github(
            title=title_full,
            details=details_console_full,
            pprint=False,
        ) if self.github_console else _pprint.entry_console(
            title=title_full,
            details=details_console_full,
            seperator_top=self._entry_seperator_top,
            seperator_bottom=self._entry_seperator_bottom,
            seperator_title=self._entry_seperator_title,
            pprint=False,
        )
        self._submit(console=console_entry, file="")
        return

    def section_end(self):
        self.section_level = max(self.section_level - 1, 1)
        return

    def _submit(self, console: str, file: str | _html.Element | _html.ElementCollection):
        console_entry = f"{console}\n"
        file_entry = f"{file}\n"
        self._log_console += console_entry
        self._log_html += file_entry
        if self.realtime_output:
            print(console, flush=True)
            if self.output_html_filepath:
                with open(self.output_html_filepath, "a") as f:
                    f.write(file_entry)
        return


def logger(
    realtime_output: bool = True,
    github_console: bool = True,
    output_html_filepath: str | _Path | None = "log.html",
    initial_section_level: _Literal[1, 2, 3, 4, 5] = 1,
    h1_kwargs: dict | None = None,
    h2_kwargs: dict | None = None,
    h3_kwargs: dict | None = None,
    h4_kwargs: dict | None = None,
    h5_kwargs: dict | None = None,
    h6_kwargs: dict | None = None,
    symbol_bulletpoint: str = "🔘",
    symbol_success: str = "✅",
    symbol_skip: str = "❎",
    symbol_error: str = "⛔",
    symbol_warning: str = "🚨",
    symbol_attention: str = "❗",
    entry_seperator_top: str = "="*35,
    entry_seperator_bottom: str = "="*35,
    entry_seperator_title: str = "-"*30,
):
    return Logger(
        realtime_output=realtime_output,
        github_console=github_console,
        output_html_filepath=output_html_filepath,
        initial_section_level=initial_section_level,
        h1_kwargs=h1_kwargs,
        h2_kwargs=h2_kwargs,
        h3_kwargs=h3_kwargs,
        h4_kwargs=h4_kwargs,
        h5_kwargs=h5_kwargs,
        h6_kwargs=h6_kwargs,
        symbol_bulletpoint=symbol_bulletpoint,
        symbol_success=symbol_success,
        symbol_skip=symbol_skip,
        symbol_error=symbol_error,
        symbol_warning=symbol_warning,
        symbol_attention=symbol_attention,
        entry_seperator_top=entry_seperator_top,
        entry_seperator_bottom=entry_seperator_bottom,
        entry_seperator_title=entry_seperator_title,
    )


import argparse
from pathlib import Path
from actionman import pprint


def h1():
    kwargs = get_args_heading()
    pprint.h1(**kwargs)
    return


def h2():
    kwargs = get_args_heading()
    pprint.h2(**kwargs)
    return


def h3():
    kwargs = get_args_heading()
    pprint.h3(**kwargs)
    return


def h4():
    kwargs = get_args_heading()
    pprint.h4(**kwargs)
    return


def h5():
    kwargs = get_args_heading()
    pprint.h5(**kwargs)
    return


def h6():
    kwargs = get_args_heading()
    pprint.h6(**kwargs)
    return


def entry_github() -> None:
    kwargs = get_args_entry_gh()
    pprint.entry_github(**kwargs)
    return


def file():
    args = get_args_file()
    content = Path(args.path).read_text()
    if args.strip:
        content = content.strip()
    margin_top = "\n" * args.margin
    margin_bottom = margin_top if content.endswith("\n") else f"{margin_top}\n"
    print(f"{margin_top}{content}{margin_bottom}", flush=True)
    return


def get_args_heading() -> dict:

    def parse_text_styles(styles: list | None):
        if styles is None:
            return
        if len(styles) == 1 and styles[0].lower() == "false":
            return False
        return [int(style) if style.isdigit() else style for style in kwargs.text_styles]

    def parse_color(color: list | None):
        if color is None:
            return
        if len(color) == 1:
            color = color[0]
            if color.lower() == "false":
                return False
            return int(color) if color.isdigit() else color
        if len(color) != 3:
            raise ValueError(f"Invalid text color: {color}")
        parsed_color = []
        for c in color:
            if c.isdigit():
                parsed_color.append(int(c))
            else:
                raise ValueError(f"Invalid text color: {color}")
        return tuple(parsed_color)

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "title",
        type=str, help="Heading title."
    )
    parser.add_argument(
        "-w", "--width",
        type=int,
        default=None,
        help="Width of the heading box in characters."
    )
    parser.add_argument(
        "-mt", "--margin-top",
        type=int,
        default=None,
        help="Number of blank lines to print before the heading box."
    )
    parser.add_argument(
        "-mb", "--margin-bottom",
        type=int,
        default=None,
        help="Number of blank lines to print after the heading box."
    )
    parser.add_argument(
        "-ts", "--text-styles",
        type=str,
        default=None,
        nargs="+",
        help=(
            "One or more text styles to apply to the heading title, "
            "each denoted either as an integer or string from the following list: "
            "0: normal, 1: bold, 2: faint, 3: italic, 4: underline, "
            "5: blink, 6: blink_fast, 7: reverse, 8: conceal, 9: strike. "
            "Set to 'false' to disable all text styles."
        )
    )
    parser.add_argument(
        "-tc", "--text-color",
        type=str,
        default=None,
        nargs="+",
        help=(
            "Text color of the heading title, "
            "denoted either as three integers representing the RGB value, "
            "or as an integer or string from the following list: "
            "30: black, 31: red, 32: green, 33: yellow, 34: blue, 35: magenta, 36: cyan, 37: white, "
            "90: b_black, 91: b_red, 92: b_green, 93: b_yellow, "
            "94: b_blue, 95: b_magenta, 96: b_cyan, 97: b_white. "
            "Set to 'false' to disable text color."
        )
    )
    parser.add_argument(
        "-bc", "--background-color",
        type=str,
        default=None,
        nargs="+",
        help=(
            "Background color of the heading title, "
            "denoted either as three integers representing the RGB value, "
            "or as an integer or string from the following list: "
            "40: black, 41: red, 42: green, 43: yellow, 44: blue, 45: magenta, 46: cyan, 47: white, "
            "100: b_black, 101: b_red, 102: b_green, 103: b_yellow, "
            "104: b_blue, 105: b_magenta, 106: b_cyan, 107: b_white. "
            "Set to 'false' to disable background color."
        )
    )

    kwargs = parser.parse_args()
    kwargs.text_styles = parse_text_styles(kwargs.text_styles)
    kwargs.text_color = parse_color(kwargs.text_color)
    kwargs.background_color = parse_color(kwargs.background_color)
    return {param: arg for param, arg in vars(kwargs).items() if arg is not None}


def get_args_entry_gh() -> dict:
    parser = argparse.ArgumentParser()
    parser.add_argument("title", type=str, help="Summary of the data.")
    parser.add_argument("details", type=str, help="Details of the data.")
    args = parser.parse_args()
    return vars(args)


def get_args_file():
    parser = argparse.ArgumentParser()
    parser.add_argument("path", type=str, help="Path to the file.")
    parser.add_argument(
        "-m", "--margin",
        type=int,
        required=False,
        default=2,
        help="Number of blank lines to add to the top and bottom of the file."
    )
    parser.add_argument(
        "-s", "--strip",
        action="store_true",
        help="Strip the file of leading and trailing whitespace before applying the margin."
    )
    return parser.parse_args()




import subprocess as _subprocess
from pathlib import Path as _Path
from typing import NamedTuple as _NamedTuple


class ShellOutput(_NamedTuple):
    cmd: str
    out: str | bytes | None = None
    err: str | bytes | None = None
    code: int | None = None

    @property
    def executed(self) -> bool:
        return self.code is not None

    @property
    def success(self) -> bool:
        return self.code == 0

    @property
    def details(self) -> tuple[str, ...]:
        details = (
            f"Command: {self.cmd}",
            f"Executed: {self.executed}",
            f"Exit Code: {self.code}",
            f"Output: {self.out or None}",
            f"Error: {self.err or None}",
        )
        return details

    @property
    def summary(self) -> str:
        if not self.executed:
            return f"Command could not be executed."
        if not self.success:
            return f"Command failed with exit code {self.code}."
        return f"Command executed successfully."


def run(
    command: list[str],
    cwd: str | _Path | None = None,
    text_output: bool = True,
) -> ShellOutput:
    cmd_str = " ".join(command)
    try:
        process = _subprocess.run(command, text=text_output, cwd=cwd, capture_output=True)
    except FileNotFoundError:
        return ShellOutput(cmd=cmd_str)
    out = process.stdout.strip() if text_output else process.stdout
    err = process.stderr.strip() if text_output else process.stderr
    code = process.returncode
    return ShellOutput(cmd=cmd_str, out=out or None, err=err or None, code=code)


def python(
    command: list[str],
    cwd: str | _Path | None = None,
    text_output: bool = True,
) -> ShellOutput | None:
    return run(command=["python", *command], cwd=cwd, text_output=text_output)


def python_module(
    command: list[str],
    cwd: str | _Path | None = None,
    text_output: bool = True,
) -> ShellOutput | None:
    return python(command=["-m", *command], cwd=cwd, text_output=text_output)


def pip(
    command: list[str],
    cwd: str | _Path | None = None,
    text_output: bool = True,
) -> ShellOutput | None:
    return python_module(command=["pip", *command], cwd=cwd, text_output=text_output)


def pip_list(
    cwd: str | _Path | None = None,
    text_output: bool = True,
) -> ShellOutput | None:
    return pip(command=["list"], cwd=cwd, text_output=text_output)


def pip_install(
    command: list[str],
    cwd: str | _Path | None = None,
    text_output: bool = True,
) -> ShellOutput | None:
    return pip(command=["install", *command], cwd=cwd, text_output=text_output)


def pip_install_requirements(
    path: str | _Path,
    cwd: str | _Path | None = None,
    text_output: bool = True,
) -> ShellOutput | None:
    return pip_install(command=["-r", str(path)], cwd=cwd, text_output=text_output)


def pip_install_package(
    name: str,
    requirement_specifier: str | None = None,
    upgrade: bool = False,
    install_dependencies: bool = True,
    index: str | None = None,
    cwd: str | _Path | None = None,
    text_output: bool = True,
) -> ShellOutput | None:
    index_name_url = {
        "testpypi": "https://test.pypi.org/simple/",
    }
    command = []
    if upgrade:
        command.append("--upgrade")
    command.append(f"{name}{requirement_specifier or ''}")
    if not install_dependencies:
        command.append("--no-deps")
    if index:
        index_url = index_name_url.get(index) or index
        command.extend(["--index-url", index_url])
    return pip_install(command=command, cwd=cwd, text_output=text_output)



from typing import Callable as _Callable, get_type_hints as _get_type_hints, Type as _Type
import os as _os
import json as _json
import inspect as _inspect
import base64 as _base64

from actionman.log import Logger as _Logger


def read_environment_variable(
    name: str,
    typ: _Type[str | bool | int | float | list | dict] = str,
    missing_ok: bool = False,
    logger: _Logger | None = None,
    log_value: bool = True,
) -> str | bool | int | float | list | dict | None:
    """
    Parse inputs from environment variables.
    """
    def log(status, summary):
        if logger:
            logger.entry(
                status=status,
                title=f"Read environment variable '{name}'",
                summary=summary,
                details=[
                    f"Name: {name}",
                    f"Type: {typ.__name__}",
                    f"Required: {'No' if missing_ok else 'Yes'}",
                    f"Value: {value if log_value else '**REDACTED**'}"
                ],
            )
        return

    def cast_error(expected_typ: str, exception: Exception | None = None):
        log(status="error", summary=f"Value could not be casted to {expected_typ}.")
        to_raise = TypeError(f"Environment variable {name} could not be casted to {expected_typ}.")
        if exception:
            raise to_raise from exception
        else:
            raise to_raise

    value = _os.environ.get(name)
    if value is None:
        if missing_ok:
            log(status="skip", summary=f"Environment variable '{name}' is not set, but it is not required.")
            return
        else:
            log(status="error", summary=f"Environment variable '{name}' is not set.")
            raise ValueError(f"Environment variable '{name}' is not set.")
    if typ is str:
        if isinstance(value, str):
            value_casted = value
        else:
            try:
                value_casted = str(value)
            except Exception as e:
                cast_error(expected_typ="string", exception=e)
    elif typ is bool:
        if isinstance(value, bool):
            value_casted = value
        elif isinstance(value, str) and value.lower() in ("true", "false", ""):
            value_casted = value.lower() == "true"
        else:
            cast_error(expected_typ="boolean")
    elif typ is int:
        if isinstance(value, int):
            value_casted = value
        elif isinstance(value, str):
            try:
                value_casted = int(value)
            except Exception as e:
                cast_error(expected_typ="integer", exception=e)
        else:
            cast_error(expected_typ="integer")
    elif typ is float:
        if isinstance(value, float):
            value_casted = value
        elif isinstance(value, str):
            try:
                value_casted = float(value)
            except Exception as e:
                cast_error(expected_typ="float", exception=e)
        else:
            cast_error(expected_typ="float")
    elif typ is list:
        if isinstance(value, list):
            value_casted = value
        elif isinstance(value, str):
            try:
                value_casted = _json.loads(value, strict=False)
            except Exception as e:
                cast_error(expected_typ="list", exception=e)
        else:
            cast_error(expected_typ="list")
    elif typ is dict:
        if isinstance(value, dict):
            value_casted = value
        elif isinstance(value, str):
            try:
                value_casted = _json.loads(value, strict=False)
            except Exception as e:
                cast_error(expected_typ="dict", exception=e)
        else:
            cast_error(expected_typ="dict")
    else:
        raise TypeError(f"The specified type '{typ}' for environment variable '{name}' is not supported.")
    log(status="success", summary=f"Environment variable '{name}' was read successfully.")
    return value_casted


def read_environment_variables(
    *variables_data: tuple[str, _Type[str | bool | int | float | list | dict], bool, bool],
    name_prefix: str = "",
    logger: _Logger | None = None,
) -> dict[str, str | bool | int | float | list | dict | None]:
    """
    Parse inputs from environment variables.
    """
    if logger:
        logger.section("Read Environment Variables")
    variables = {}
    for name, typ, missing_ok, log_value in variables_data:
        variables[name] = read_environment_variable(
            name=f"{name_prefix}{name}",
            typ=typ,
            missing_ok=missing_ok,
            logger=logger,
            log_value=log_value,
        )
    if logger:
        logger.section_end()
    return variables


def read_function_args_from_environment_variables(
    function: _Callable,
    name_prefix: str = "",
    hide_args: tuple[str, ...] | list[str] = tuple(),
    ignore_params: tuple[str, ...] | list[str] = tuple(),
    logger: _Logger | None = None,
) -> dict[str, str | bool | int | float | list | dict | None]:
    """
    Parse inputs from environment variables.
    """
    if logger:
        logger.section("Read Input Arguments")
    default_args = {
        k: v.default for k, v in _inspect.signature(function).parameters.items()
        if v.default is not _inspect.Parameter.empty
    }
    params = _get_type_hints(function)
    params.pop("return", None)
    args = {}
    for name, typ in params.items():
        if name not in ignore_params:
            arg = read_environment_variable(
                name=f"{name_prefix}{name}".upper(),
                typ=typ,
                missing_ok=name in default_args,
                logger=logger,
                log_value=name not in hide_args,
            )
            args[name] = arg if arg is not None else default_args[name]
    if logger:
        logger.section_end()
    return args


def write_github_outputs(kwargs: dict, env: bool = False, logger: _Logger | None = None) -> None:

    def format_output(var_name, var_value) -> str | None:
        if isinstance(var_value, str):
            if "\n" in var_value:
                with open("/dev/urandom", "rb") as f:
                    random_bytes = f.read(15)
                random_delimeter = _base64.b64encode(random_bytes).decode("utf-8")
                return f"{var_name}<<{random_delimeter}\n{var_value}\n{random_delimeter}"
        elif isinstance(var_value, (dict, list, tuple, bool, int)):
            var_value = _json.dumps(var_value)
        else:
            return
        return f"{var_name}={var_value}"
    if logger:
        logger.section(f"Write {'Environment Variables' if env else 'Step Outputs'}")
    with open(_os.environ["GITHUB_ENV" if env else "GITHUB_OUTPUT"], "a") as fh:
        for idx, (name, value) in enumerate(kwargs.items()):
            name_formatted = name.replace("_", "-") if not env else name.upper()
            title = f"Write '{name_formatted}' ({type(value).__name__})"
            value_formatted = format_output(name_formatted, value)
            if not value_formatted:
                summary = f"Invalid type {type(value)} for variable '{name_formatted}': {value}."
                if logger:
                    logger.entry(
                        status="error",
                        title=title,
                        summary=summary,
                    )
                raise TypeError(f"Failed to write output variable. {summary}")
            print(value_formatted, file=fh)
            if logger:
                logger.entry(
                    status="success",
                    title=title,
                    summary=(
                        f"Output variable '{name_formatted}' ({type(value).__name__}) "
                        f"was successfully written to {'environment' if env else 'step'} outputs."
                    ),
                )
    if logger:
        logger.section_end()
    return


def write_github_summary(content: str, logger: _Logger | None = None) -> None:
    with open(_os.environ["GITHUB_STEP_SUMMARY"], "a") as fh:
        print(content, file=fh)
    if logger:
        logger.entry(
            status="success",
            title="Write Job Summary",
            summary=(
                f"Job summary ({len(content)} chars) was successfully written "
                f"to 'GITHUB_STEP_SUMMARY' environment variable."
            ),
        )
    return
